require 'faker'
require 'soar/registry/identity'
require 'soar/registry/directory'
require 'soar/authentication/identity_uuid_translator/provider/customer'
require 'soar/authentication/identity_uuid_translator/uuid_generator'

module Soar
  module Registry
    module Identity
      module Test
        module OrchestrationProvider
          module Customer
            class Uuid

              def initialize
                Faker::UniqueGenerator.clear
                @identity = {
                  "ID" => Faker::Number.unique.number(4).to_i,
                  "First_Name" => Faker::Name.unique.first_name,
                  "Surname" => Faker::Name.unique.last_name,
                  "Notifyemail_Invoice" => "#{Faker::Internet.unique.email}",
                  "Client_Number" => "C#{Faker::Number.unique.number(10)}"
                }
                @identity_uuid = Soar::Authentication::IdentityUuidTranslator::UuidGenerator.generate("#{Soar::Authentication::IdentityUuidTranslator::Provider::Customer::PREFIX}#{@identity['ID']}")
              end

              def given_identity_directory
                wait_for_database do
                  @identity_directory_config = YAML.load_file("config/#{ENV['CUSTOMER_DIRECTORY_CONFIG_FILE']}")
                  @identity_directory =  Soar::Registry::Directory.new(
                    Soar::Registry::Directory::Provider::Mysql.new(@identity_directory_config['config'].map { |k, v| [k.to_sym, v] }.to_h)
                  )
                end
              end

              def given_roles_directory
                @roles_directory_config = YAML.load_file("config/#{ENV['ROLES_DIRECTORY_CONFIG_FILE']}")
                @roles_directory = Soar::Registry::Directory.new(
                  Soar::Registry::Directory::Provider::DynamoDb.new(@roles_directory_config['config'].map { |k, v| [k.to_sym, v]}.to_h)
                )
              end

              def given_roles
                @roles = generate_roles()
                wait_for_database {
                  create_roles_table(@roles_directory, @roles_directory_config)
                  populate_roles_table(@roles_directory, @roles, @identity_uuid)
                }
              end

              def given_role_with_attributes
                @roles = generate_roles()
                @roles[0]['attributes'] = {
                  'profiles' => [@identity['Client_Number'], "C#{Faker::Number.unique.number(10)}"]
                }
                wait_for_database {
                  create_roles_table(@roles_directory, @roles_directory_config)
                  populate_roles_table(@roles_directory, @roles, @identity_uuid)
                }
              end

              def given_identity
                wait_for_database do
                  create_identity_table(@identity_directory)
                  @identity_directory.put(@identity)
                end
              end

              def given_identity_registry
                identity_provider = Soar::Registry::Identity::Provider::Customer::Uuid.new({
                  identity_directory: Soar::Registry::Identity::Directory.new({
                    directory: @identity_directory,
                    fetch_index: 'ID',
                    search_index: 'Notifyemail_Invoice'
                  }),
                  role_directory: Soar::Registry::Identity::Directory.new({
                    directory: @roles_directory,
                    fetch_index: ['identity_uuid', 'identity_role'],
                    search_index: 'identity_uuid'
                  })
                })

                @idr = Soar::Registry::Identity.new(identity_provider)
              end

              ##### when

              def get_roles
                @result = @idr.get_roles(@identity_uuid) 
              end

              def get_role_attributes
                @result = @idr.get_attributes(@identity_uuid, @roles.nil? ? Faker::Company.unique.profession : @roles[0]['role'])
              end

              def get_identity_attributes
                begin
                  @result = @idr.get_attributes(@identity_uuid)
                rescue SoarIdm::IdentityError => e
                  @error = e
                end
              end

              ##### then

              def identity_attributes?
                roles = {}
                @roles.each do |entry| 
                  roles[entry['role']] = {}
                end if not @roles.nil?
                expectation = {
                  "identity_uuid" => @identity_uuid,
                  "firstname" => @identity['First_Name'],
                  "lastname" => @identity['Surname'], 
                  "email" => @identity['Notifyemail_Invoice'],
                  "roles" => roles
                }
                @result == expectation
              end

              def nil?
                @result == nil
              end
              
              def role_with_empty_attributes?
                @result == {
                  @roles[0]['role'] => {}
                }
              end

              def role_with_attributes?
                role_with_attributes = {
                  @roles[0]['role'] => Hashie.stringify_keys(@roles[0]['attributes'])
                }
                @result == role_with_attributes
              end

              def roles?
                roles = @roles.map do |entry| 
                  entry['role']
                end
                @result.sort == roles.sort
              end

              def no_roles?
                @result == []
              end

              def identity_error?
                @error.is_a?(SoarIdm::IdentityError)
              end

              private

              def generate_roles()
                return [{
                  'source' => "#{Soar::Authentication::IdentityUuidTranslator::Provider::Customer::PREFIX}#{@identity['ID']}",
                  'role' => Soar::Authentication::IdentityUuidTranslator::Provider::Customer::ROLE
                },{
                  'role' => Faker::Company.unique.profession
                }]
              end

              def populate_roles_table(roles_directory, roles, identity_uuid)
                  roles.each { |role|
                    entry = {
                      "identity_uuid" => identity_uuid,
                      "identity_role" => role['role']
                    }
                    entry['identity_source'] = role['source'] if role.key?('source')
                    entry['identity_role_attributes'] = role['attributes'] if role.key?('attributes')
                    roles_directory.put(entry)
                  }
              end

              def create_roles_table(roles_directory, roles_directory_config)
                roles_directory.provider.recreate_table({
                  name: roles_directory_config['config']['table']['name'],
                  structure: JSON.parse(File.read("lib/soar/registry/identity/test/fixtures/roles_table.json"))
                })
              end

              def create_identity_table(identity_directory)
                identity_directory.provider.client.query("DROP TABLE IF EXISTS `Client`")
                identity_directory.provider.client.query("CREATE TABLE Client ( ID int(11) NOT NULL AUTO_INCREMENT, Client_Number varchar(15) NOT NULL DEFAULT '', First_Name varchar(70) DEFAULT NULL, Surname varchar(70) DEFAULT NULL, Notifyemail_Invoice text, PRIMARY KEY (ID), KEY Notifyemail_Invoice (Notifyemail_Invoice(20)), KEY Client_Number (Client_Number))")
              end

              def deep_copy(o)
                Marshal.load(Marshal.dump(o))
              end

              def wait_for_database
                return 10.times do |i|
                  begin
                    break yield
                  rescue Soar::Registry::Directory::Error::NetworkingError, Net::LDAP::Error => e
                    sleep(10)
                  end
                end
              end

            end
          end
        end
      end
    end
  end
end
